(*
簡易HTMLパーサ
XMLとして破綻しているHTMLをXMLパーサを通せる程度に構造化する。
入力はUTF-8を想定
OCaml 4.02
2015.02.11 - 2015.04.04 Takeshi Yoneki
2020.07.05 ロイター向けに改変
2021.02.27 東京新聞のバグ対処
2021.08.05 テレビ王国の間違った閉じタグ
2021.08.10 DOCTYPEを出力できるようにするため、Document型を用意
2021.08.12 整形出力機能
2021.08.31 終了タグ判定関連の改変
2021.09.05 シングルクオート対応など
ostra@asahi.email.ne.jp
*)

open TTools

exception ERROR_PARSE
exception ERROR_BUG of string

type t = 
  | Element of (string * (string * string) list * t list)
  | PCData of string
  | CData of string
  | Comment of string
  | Doctype of string
  | Decl of string
  | Document of t list
  | Empty

type partKind = PK_Start | PK_End | PK_Empty | PK_Text | PK_Comment | PK_Doctype | PK_Decl | PK_CData

module StringSet = Set.Make(String)

let isSpace ch =
  ch = ' ' || ch = '\t' || ch = '\n' || ch = '\r'

let parsePhase1 data =
  let part = ref ([]: char list) in
  let pool = ref ([]: (partKind * string) list) in
  let inTag = ref false in
  let inQuote = ref '\x00' in
  let inAttrValue = ref false in
  let inComment = ref false in
  let inCData = ref false in
  let inScript = ref false in
  let startLessThan = ref false in
  (**)
  let addPartToPool ps =
    if !part <> [] then (
      let row =
       match ps with
       | Some p -> p
       | None -> TCharList.to_string_rev !part
      in
      (*let row = TCharList.to_string_rev !part in*)
      let s = String.lowercase row in
      (*print_endline s ;*)
      let k = 
        if s.[0] = '<' && List.nth !part 0 = '>' then (
          (* ここに来る場合、文字列長は必ず2以上なのでs.[1]は存在する。 *)
          if s.[1] = '/' then (
            PK_End
          )
          else if s.[1] = '!' then (
            if TString.starts_with s "<!doctype" then
              PK_Doctype
            else if TString.starts_with s "<![cdata[" && TString.ends_with s "]]>" then
              PK_CData
            else if TString.starts_with s "<!--" && TString.ends_with s "-->" then
              PK_Comment
            else          
              PK_Start
          )
          (* ここに来る場合、キャラクタ数は必ず2以上なので(!part).[1]は存在する。 *)
          else if List.nth !part 1 = '/' then
            PK_Empty
          else if s.[1] = '?' then (
            if List.nth !part 1 = '?' then
              PK_Decl
            else
              PK_Start
          )
          else
            PK_Start
        )
        else
          PK_Text
      in
      pool := (k, row) :: !pool;
      part := []
    ) ;
    startLessThan := false
  in
  String.iter
    (fun ch ->
      if ch = '\r' then
        ()
      else if !inCData then (
        part := ch :: !part;
        if ch = '>' then (
          let p = TCharList.to_string_rev !part in
          if TString.ends_with p "]]>" then (
            addPartToPool (Some p);
            inCData := false ;
            inTag := false ;
            inAttrValue := false ;
          )
        )
      )
      else if !inComment then (
        part := ch :: !part;
        if ch = '>' then (
          let p = TCharList.to_string_rev !part in
          if TString.ends_with p "-->" then (
            addPartToPool (Some p);
            inComment := false ;
            inTag := false ;
            inAttrValue := false ;
          )
        )
      )
      else if !inScript then (
        if ch = '<' then (
          if !part <> [] then (
            pool := (PK_Text, TCharList.to_string_rev !part) :: !pool;
            part := [] ;
          ) ;
          startLessThan := true
        );
        part := ch :: !part ;
        if !startLessThan && (ch = '>' || ch = '-' || ch = '[') then (
          let p = TCharList.to_string_rev !part in
          let q = String.lowercase p in
          if ch = '>' && q = "</script>" then (
            addPartToPool (Some p) ;
            inScript := false ;
            inTag := false ;
            inAttrValue := false ;
          )
          else if ch = '-' && q = "<!--" then (
            inComment := true ;
          )
          else if ch = '[' && q = "<![cdata[" then (
            inCData := true ;
          )
        )
      )
      else if !inTag then (
        if !inQuote <> '\x00' then (
          if ch = !inQuote then (
            inQuote := '\x00' ;
            inAttrValue := false ;
          );
          part := ch :: !part
        )
        else if ch = '>' then (
          part := ch :: !part;
          let p = TCharList.to_string_rev !part in
          (* p.[0] = '<'ならばキャラクタ数は必ず2以上なので(!part).[1]は存在する。*)
          if p.[0] = '<' && List.nth !part 1 <> '/' then (
            let q = String.lowercase p in
            if TString.starts_with q "<script " || q = "<script>" then (
              inScript := true
            ) ;
          ) ;
          addPartToPool (Some p) ;
          inTag := false ;
          inAttrValue := false ;
        )
        else if !inAttrValue && isSpace ch then (
          (* 「=」後の空白を詰める。属性値のとき空白前の文字が「=」でなければ属性値終了。
           タグの中なら必ず(!part).[0]は存在する。 *)
          if List.hd !part = '=' then
            ()
          else (
            inAttrValue := false ;
            part := ch :: !part ;
          )
        )
        else if ch = '=' then (
            inAttrValue := true ;
            part := ch :: !part ;
        )
        else if ch = '\"' || ch = '\'' then (
          (* 東京新聞で間違ったダブルクオートが混入している。
          タグの中でのクオート終了直後のクオート開始は無視する。
          2021.02.27 *)
          if not !inAttrValue then (
            ()
          )
          else (
            inQuote := ch ;
            part := ch :: !part ;
          )
        )
        else (
          part := ch :: !part
        ) ;
        if ch = '-' then (
          let p = TCharList.to_string_rev !part in
          if p = "<!--" then (
            inComment := true ;
          )
        )
        else if ch = '[' then (
          let p = String.lowercase (TCharList.to_string_rev !part) in
          if p = "<![cdata[" then (
            inCData := true ;
          )
        )
      )
      else (
        if ch = '<' then (
          addPartToPool None ;
          inTag := true ;
        ) ;
        part := ch :: !part ;
      )
    )
    data ;
  if !part <> [] then
    addPartToPool None ;
  List.rev !pool

let fixCharEntity tx =
  (* 実体参照は文字にする *)
  let rec loop t ac =
    let (n, p) = TCharEntity.find_char_entity t in
    if n < 0 then
      String.concat "" (List.rev (t :: ac))
    else
      let tl = String.length t in
      let pl = String.length p in
      let ac = (TCharEntity.decode_char_entity p) :: (String.sub t 0 n) :: ac in
      let tr = String.sub t (n + pl) (tl - (n + pl)) in
      loop tr ac
  in
  loop tx []

let parseStart k row =
  let part = ref ([]: char list) in
  let pool = ref ([]: (string * string) list) in
  let inQuote = ref '\x00' in
  let name = ref "" in
  let inValue = ref false in
  let sub =
    if k = PK_Empty then
      (String.sub row 1 ((String.length row) - 3)) ^ " "
    else
      (String.sub row 1 ((String.length row) - 2)) ^ " "
  in
  let enter force =
    if !part <> [] || force then (
      let row = TCharList.to_string_rev !part in
      part := [] ;
      if !inValue then (
        pool := (!name, row) :: !pool ;
        name := ""
      )
      else (
        if !name <> "" then (
          pool := (!name, !name) :: !pool ;
        ) ;
        name := row
      )
    )
  in
  String.iter
    (fun ch ->
      if !inQuote <> '\x00' then (
        if ch = !inQuote then (
          (* Quoteを閉じるときは空の値も有効。 *)
          enter true ;
          inQuote := '\x00' ;
          inValue := false ;
        )
        else (
          part := ch :: !part
        )
      )
      else (
        if ch = '\"' || ch = '\'' then (
          enter false ;
          inQuote := ch
        )
        else if ch = '=' then (
          enter false ;
          inValue := true ;
        )
        else if isSpace ch then (
          enter false ;
          inValue := false ;
        )
        else (
          part := ch :: !part
        )
      )
    )
    sub;
  enter false ;
  if !name <> "" then (
    pool := (!name, !name) :: !pool ;
  ) ;
  let px = List.rev !pool in
  if px <> [] then
    let (tag, _) = List.hd px in
    let attr =
      List.map 
        (fun (n, v) ->
          (n, fixCharEntity v)
        )
        (List.tl px)
    in
    Element (tag, attr, [])
  else
    Empty
  
let parseEnd row =
  Element (String.trim (String.sub row 2 ((String.length row) - 3)), [], [])

let fixComment t =
  (* コメント内に"--"を記載してしまうページがある *)
  let s = String.sub t 4 ((String.length t) - 7) in
  let s = TString.replace s "--" "&#x2d;&#x2d;" in
  s

let parsePhase2 data =
  List.map
    (fun (k, row) ->
      let part =
        if k = PK_Start || k = PK_Empty then
          parseStart k row
        else if k = PK_End then
          parseEnd row
        else if k = PK_Text then
          PCData (fixCharEntity row)
        else if k = PK_CData then
          CData (String.sub row 9 ((String.length row) - 12))
        else if k = PK_Comment then
          Comment (fixComment row)
        else if k = PK_Doctype then
          Doctype row
        else if k = PK_Decl then
          Decl row
        else
          PCData (row)
      in
      (k, part)
    )
    data

(*let print_stack ls =
  print_endline (String.concat " " ls)*)

(* 末尾再帰にできそうにないのでループにする *)
let parsePhase3 data =
  (* 終了タグの名前のセット *)
  let endTags =
    StringSet.of_list
      (List.map
        (fun (_, p) -> 
          match p with
          | Element (tag, _, _) ->
            String.lowercase tag
          | _ -> 
            raise (ERROR_BUG "1")
        )
        (List.filter (fun (k, _) -> k = PK_End) data))
  in
  let hasEnd1 tag =
    StringSet.mem tag endTags
  in
  (* タグに囲まれたデータを分離する *)
  let mkInnerData tn data =
    let dx = ref data in
    let bracket = ref 1 in
    let inner = ref [] in
    while !bracket > 0 && !dx <> [] do
      let d = List.hd !dx in
      inner := d :: !inner ;
      dx := List.tl !dx ;
      let (k, p) = d in
      let t = 
        match p with
        | Element (tag, _, _) -> String.lowercase tag
        | _ -> ""
      in
      if k = PK_Start && t = tn then
        bracket := !bracket + 1
      else if k = PK_End && t = tn then
        bracket := !bracket - 1
      else
        ()
    done ;
    (!dx, List.rev !inner, !bracket = 0)
  in
  (**)
  let stack = ref [] in
  (**)
  let rec mkTree data =
    let dx = ref data in
    let brk = ref false in
    let sibling = ref [] in
    let q = ref "" in
    while !dx <> [] && not !brk do
      let (k, p) = List.hd !dx in
      dx := List.tl !dx;
      match k with
      | PK_Start -> (
        match p with
        | Element (tag, attr, _) ->
          let tn = String.lowercase tag in
          if hasEnd1 tn && !dx <> [] then
            let (rest1, inner, hasEnd2) = mkInnerData tn !dx in
            if hasEnd2 then (
              stack := tn :: !stack ;
              let (rest2, child, cue) = mkTree inner in
              dx := List.append rest2 rest1 ;
              stack := List.tl !stack ;
              let e = Element (tag, attr, List.rev child) in
              sibling := e :: !sibling ;
              if !stack <> [] && cue <> tn && List.mem cue !stack then (
                  q := cue ;
                  brk := true
              )
            )
            else (
              sibling := p :: !sibling
            )
          else (
            sibling := p :: !sibling
          )
        | _ -> 
          ()
          (*raise (ERROR_BUG "2")*)
      )
      | PK_End -> (
        match p with
        | Element (tag, _, _) -> (
          let tn = String.lowercase tag in
          let currentTag =
            if !stack <> [] then 
              List.hd !stack 
            else 
              ""
          in
          (* テレビ王国の間違った閉じタグ 2021.08.05 *)
          if tn <> currentTag then (
            let e = Comment ("/" ^ tag) in
            sibling := e :: !sibling
          )
          else (
            q := tn ;
            brk := true
          )
        )
        | _ -> 
          raise (ERROR_BUG "3")
      )
      | _ ->
      (*| PK_Text | PK_CData | PK_Empty | PK_Comment ->*)
        sibling := p :: !sibling
        (* DOCTYPEなどは捨てる -> 捨てるのをやめる *)
    done ;
    (!dx, !sibling, !q)
  in
  (**)
  let (_, s, _) = mkTree data in
  List.rev s

let print_htmlp p =
  let rec subfn p sp =
    match p with
    | Element (tag, attr, child) ->
      print_endline (sp ^ tag);
      List.iter
        (fun (n, v) ->
          print_endline (sp ^ n ^ "=" ^ v)
        )
        attr;
      List.iter
        (fun c ->
          subfn c (sp ^ " ")
        )
        child
    | PCData (text) ->
      let t = String.trim text in
      if t <> "" then
        print_endline (sp ^ t)
      else
        ()
    | _ ->
      ()
  in
  subfn p ""

let print_phase2 p =
  List.iter
    (fun (k, part) ->
      print_htmlp part
    )
    p

let print_phase3 p =
  List.iter
    (fun e ->
      print_htmlp e
    )
    p

let parse data =
  let p1 = parsePhase1 data in
  let p2 = parsePhase2 p1 in
  (*print_phase2 p2 ;*)
  let p3 = parsePhase3 p2 in
  (*print_phase3 p3 ;*)
  Document p3

let toXmlString node =
  (* 実体参照にすべき文字は実体参照にする *)
  let toCharEntity isAttr t =
    let t = TString.replace t "&" "&amp;" in
    let t = TString.replace t ">" "&gt;" in
    let t = TString.replace t "<" "&lt;" in
    let t = 
      if isAttr then
        (* 2023.01.25 属性値の場合に&amp;などを加える。 *)
        let t = TString.replace t "\"" "&quot;" in
        let t = TString.replace t "'" "&quos;" in
        t
      else
        t
    in
    t
  in
  let fixAttrValue t =
    toCharEntity true t
  in
  let fixText t =
    toCharEntity false t
  in
  let rec onNode e acc =
    match e with
    | Element (tag, attr, child) ->
      let axs = 
        List.map
          (fun (n, v) ->
            n ^ "=\"" ^ (fixAttrValue v) ^ "\""
          )
          attr
      in
      let ax = String.concat " " axs in
      let s = "<" ^ tag ^ 
        if ax = "" then
          ""
        else 
          " " ^ ax
      in
      if child = [] then
        (s ^ " />") :: acc
      else
        let acc = (s ^ ">") :: acc in
        let acc = 
          List.fold_left
            (fun acc e ->
              onNode e acc
            )
            acc
            child
        in
        ("</" ^ tag ^ ">") :: acc
    | Comment (t) ->
      ("<!--" ^ t ^ "-->") :: acc
    | PCData (t) ->
      (fixText t) :: acc
    | CData (t) ->
      (fixText t) :: acc
    | Doctype (t) ->
      t :: acc
    | Decl (t) ->
      t :: acc
    | Document (child) ->
      let acc = 
        List.fold_left
          (fun acc e ->
            onNode e acc
          )
          acc
          child
      in
      acc
    | _ ->
      acc
  in
  let acc =
    match node with
    | Document (_) -> []
    | _ -> ["\n"]
  in
  String.concat "" (List.rev (onNode node acc))

let isElement node =
  match node with
  | Element _ -> true
  | _ -> false

let isText node =
  match node with
  | PCData _ -> true
  | CData _ -> true
  | _ -> false

let setAttribute elem name value =
  match elem with
  | Element (tag, attr, child) ->
    let ax = List.filter (fun (n, v) -> n <> name) attr in
    let ax = List.rev ax in
    let ax = (name, value) :: ax in
    let ax = List.rev ax in
    Element (tag, ax, child)
  | _ ->
    elem  

let removeAttribute elem name =
  match elem with
  | Element (tag, attr, child) ->
    let ax = List.filter (fun (n, v) -> n <> name) attr in
    Element (tag, ax, child)
  | _ ->
    elem  

let getAttribute elem name =
  match elem with
  | Element (_, attr, _) ->
    let ax = List.filter (fun (n, v) -> n = name) attr in
    if ax = [] then
      ""
    else
      let (_, v) = List.hd ax in
      v
  | _ -> 
    ""

let getDocumentElement doc =
  match doc with
  | Document (child) ->
    let es = 
      List.filter
        (fun n ->
          match n with
          | Element (_, _, _) ->
            true
          | _ ->
            false
        )
        child
    in
    if es = [] then
      Empty
    else
      List.hd es
  | Element (_, _, _) ->
    doc
  | _ ->
    Empty

let getDataText node =
  match node with
  | PCData t -> t
  | CData t -> t
  | _ -> ""

let getElementText elem =
  match elem with
  | Element (_, _, child) ->
    let ts =
      List.map
        (fun n ->
          getDataText n
        )
        child
    in
    let tx = String.concat "" ts in
    tx
  | _ -> 
    ""

let getTextContent node =
  match node with
  | Element (_, _, _) ->
    getElementText node
  | _ ->
    getDataText node

let hasText elem =
  getTextContent elem <> ""

let rec getTextContentRecur node =
  match node with
  | Element (_, _, child) ->
    List.fold_left
      (fun acc n ->
        acc ^ (getTextContentRecur n)
      )
      ""
      child
  | _ ->
    getDataText node

let rec getElementsByTagName node name =
    match node with
    | Element (tag, _, child) ->
      let r1 =
        List.map
          (fun n ->
            if isElement n then
              getElementsByTagName n name
            else
              []
          )
          child
      in
      let r2 = List.concat r1 in
      let r3 =
        if tag = name || List.hd (List.rev (TString.split [":"] tag)) = name then
          List.rev (node :: (List.rev r2))
        else
          r2
      in
      r3
    | Document(child) ->
      let r =
        List.map
          (fun n ->
            if isElement n then
              getElementsByTagName n name
            else
              []
          )
          child
      in
      List.concat r
    | _ ->
      []

let getChildren elem =
  match elem with
  | Element (_, _, child) -> 
    child
  | _ -> 
    []

let getTagName elem =
  match elem with
  | Element (tag, _, _) -> 
    tag
  | _ -> 
    ""

let rec ruleTransformer trans node =
  let rec ruleTransformerSub trans node path =
    let p = 
      match node with
      | Element (tag, _, _) ->
        path ^ "/" ^ (getTagName node)
      | _ ->
        path
    in
    let r = trans node p in
    match r with
    | Element (tag, attr, child) ->
      let ns = 
        List.map 
          (fun n -> 
            ruleTransformerSub trans n p
          ) 
          child 
      in
      Element (tag, attr, ns)
    | Document (child) ->
      let ns = 
        List.map 
          (fun n -> 
            ruleTransformerSub trans n p
          ) 
          child
      in
      Document (ns)
    | _ ->
      r
  in
  ruleTransformerSub trans node ""

let pretty tab doc =
  let isSpaceString s =
    let t =
      TString.filter
        (fun ch ->
          not (isSpace ch)
        )
        s
    in
    t = ""
  in
  let hasTextPCData child =
    let child2 =
      List.filter 
        (fun e ->
          match e with
          | PCData (t) ->
            not (isSpaceString t)
          | CData (t) ->
            not (isSpaceString t)
          | _ ->
            false
        )
        child
    in
    child2 <> []
  in
  let deleteSpace doc =
    let deleteSpacePCData child =
      if child = [] || hasTextPCData child then
        child
      else
        List.filter 
          (fun n ->
            match n with
            | PCData (t) ->
              not (isSpaceString t)
            | CData (t) ->
              not (isSpaceString t)
            | _ ->
              true
          )
          child
    in
    let doc2 =
      ruleTransformer 
        (fun node path ->
          match node with
          | Element (tag, attr, child) ->
            Element (tag, attr, deleteSpacePCData child)
          | Document (child) ->
            Document (deleteSpacePCData child)
          | _ ->
            node
        )
        doc
    in
    doc2
  in
  let indentSpace doc =
    let calcIndentCount path =
      let t =
        TString.filter
          (fun ch ->
            ch = '/'
          )
          path
      in
      String.length t
    in
    let mkIndentPCData tab count =
      if count < 1 then
        PCData ""
      else
        let ts = 
          List.fold_left
            (fun acc n ->
              tab :: acc
            )
            []
            (TList.range count)
        in
        PCData (String.concat "" ts)
    in
    let endline = PCData "\n" in
    let doc2 =
      ruleTransformer 
        (fun node path ->
          match node with
          | Element (tag, attr, child) ->
            if child = [] || hasTextPCData child then
              node
            else
              let indentCount = calcIndentCount path in
              let indent = mkIndentPCData tab indentCount in
              let indentLast = mkIndentPCData tab (indentCount - 1) in
              let child2 = 
                List.fold_left
                  (fun ac e ->
                    endline :: e :: indent :: ac
                  )
                  [endline]
                  child
              in
              let child3 = indentLast :: child2 in
              Element (tag, attr, List.rev child3)
          | Document (child) ->
              let child2 = 
                List.fold_left
                  (fun ac e ->
                    endline :: e :: ac
                  )
                  []
                  child
              in
              Document (List.rev child2)
          | _ ->
            node
        )
        doc
    in
    doc2
  in
  let doc2 = deleteSpace doc in
  indentSpace doc2
